#
#  CachesDataSource.rb
#  GPSMinna
#
#  Created by Heiko on 06.09.08.
#  Copyright (c) 2008 Heiko Kretschmer. All rights reserved.
#
# Basing on file:///Developer/Documentation/DocSets/com.apple.ADC_Reference_Library.CoreReference.docset/Contents/Resources/Documents/documentation/Cocoa/Reference/ApplicationKit/Protocols/NSTableDataSource_Protocol/Reference/Reference.html

require 'osx/cocoa' # fuer RubyCocoa
require 'Date'		# fuer DateTime
require 'fileutils' # fuer File.exectable and friends
require 'Defines.rb'
require 'RobAnybody.rb' # Require here, so that it can be included below

####################################
# Includes (of Classes)            #
include		OSX
include		RobAnybody

####################################
# Klasse "CachesDataSource"        #
class CachesDataSource < OSX::NSObject
#class AppController < OSX::NSObject

####################################
# Globals                          #
@_array		= nil	# Diese Variablen nur via Accessor Methods lesen/schreiben. Danke :)
@_lock		= nil	# Diese Variablen nur via Accessor Methods lesen/schreiben. Danke :)


# Outlets
ib_outlets :_tableView

# Fill Array
def fillArray
	# Alten Inhalt kicken
	self.removeAllObjects()
	
	# Locken
	self.lock()
	
	# Lesen des Verzeichnisses ins Array
	Say("Reading GPX files...")
	thisBundle	= OSX::NSBundle.mainBundle
	username	= OSX::NSUserName()
	formatPathToFolder=FOLDERGPX #/*.{GPX,gpx}" # GPX-Ordner definieren und expanden -> Das hier vielleicht Customizebar machen? 
	@pathToGPXFolder= OSX::NSMutableString.alloc.initWithString("")
	@pathToGPXFolder.appendFormat(formatPathToFolder, username)
	Say("pathToGPXFolder: %s", @pathToGPXFolder.to_s)
#	gpxfilesarray=Dir[@pathToGPXFolder.to_s] # Den Verzeichnisinhalt in ein Ruby-Array schreiben
	gpxfilesarray=Dir.entries(@pathToGPXFolder.to_s) # Den Verzeichnisinhalt in ein Ruby-Array schreiben
	for gpxfilename in gpxfilesarray
		#puts("Processing file: " + gpxfile)
		# Testen, ob die Files KEIN "." am Anfang und dafür ein "GPX" oder "gpx" am Ende haben!
		if (gpxfilename =~ /^\w+/) then # .-Files filtern, nur GPX ueberlassen
		# works: (gpxfile =~ /^\w+/)
			gpxfilecomparray=gpxfilename.split('.')
			if(gpxfilecomparray.length == 2) then
				cache_id = gpxfilecomparray[0].upcase # Die ID ist der Cachename vor dem .gpx
				suffix   = gpxfilecomparray[1].downcase 
				if (suffix == "gpx") then
					Say("Processing file %s", gpxfilename)
				# Das Dict erzeugen lassen
				@dict=OSX::NSMutableDictionary.dictionary()
					self.fillDict(@dict, @pathToGPXFolder.to_s, gpxfilename)
					@_array.addObject(@dict)
				@dict.release()
				else
					Say("%s might also not be an gpx. Skipping.", gpxfilename)
				end # if .gpx
			else
				Say("%s contains more than one suffix. Skipping.", gpxfilename)
			end
		else
			Say("%s might not be an gpx. Skipping.", gpxfilename)
		end # if word-character
	end

	# Unlocken
	self.unlock()
end # fillArray


# Das Dict für einen Array-Eintrag (=1 Cache) fuellen
def fillDict(dict, gpxfilepath, gpxfilename)

	# GPX-File einlesen
	filepath=gpxfilepath + '/' + gpxfilename
	Say("Reading %s", filepath)
	@dict.setObject_forKey(gpxfilename, "Filename")
	filecontentarray=open(filepath) { |f| f.readlines }
	filecontentstring=filecontentarray.join(' ')
	#Say("filecontentstring: %s", filecontentstring)
	
	# GPX-File parsen
		# Checken, ob das ein Pocket Query (PQ) war...
		gpx=oneTagContentOfXMLString(filecontentstring, '<gpx ', '</gpx>') # Das zieht das gesamte GPX - leider.
		pqCreator=oneTagContentOfXMLString(filecontentstring, 'creator="', '"') # Das sollte "Groundspeak Pocket Query" liefern, wenn es ein PQ ist!
		pqIssuingDate=oneTagContentOfXMLString(filecontentstring, '<time>', '</time>') # Das sollte "2010-09-24T10:17:53.7778815Z" liefern, wenn es ein PQ ist!
		pqIssuingDateArray=pqIssuingDate.split('T')
		pqIssuingTimeArray=pqIssuingDateArray[1].split('.')
		pqIssuingDate = pqIssuingDateArray[0] + ", " + pqIssuingTimeArray[0]
		pqName=oneTagContentOfXMLString(filecontentstring, '<name>', '</name>') # Das sollte den Namen des PQ liefern, wenn es ein PQ ist!
		if (pqName) then
			Say("pqName: " + pqName)
		else	
			Say("pqName is nil!")
		end
		pqAuthor=oneTagContentOfXMLString(filecontentstring, '<author>', '</author>') # Das sollte "Groundspeak" liefern, wenn es ein PQ ist!
		if (pqCreator == "Groundspeak Pocket Query") then
			Say("Found PQ!")
			Say("pqCreator: %s", pqCreator)
			Say("pqName: %s", pqName)
			@dict.setObject_forKey(gpxfilename, "Name")
			@dict.setObject_forKey(pqCreator, "Type")
			@dict.setObject_forKey(pqIssuingDate, "Issuing Date")
			gpxwpts=tagContentsOfXMLString(filecontentstring, '<wpt ', '</wpt>') # Das zieht die gesamte Beschreibung (inkl. Sub-Tags) aus dem GPX
			Say("Found %i Waypoints in GPX!", gpxwpts.length)
			@dict.setObject_forKey(pqName + " (" + gpxwpts.length.to_s + ")", "Description")
			@dict.setObject_forKey("n/a", "Cache-ID")
			@dict.setObject_forKey("n/a", "Difficulty")
			@dict.setObject_forKey("n/a", "Terrain")
			@dict.setObject_forKey("n/a", "URL")
			@dict.setObject_forKey("n/a", "Size")
			@dict.setObject_forKey("n/a", "Country")
			@dict.setObject_forKey("n/a", "State")
			@dict.setObject_forKey("n/a", "Status")
			@dict.setObject_forKey("n/a", "Owner")
			@dict.setObject_forKey("n/a", "Found")
			# Anzahl der WPTs im GPX...
		else # else if !PQ
			Say("Found ordinary GPX! (Either a single GPX with one cache description or a Current.gpx imported from device.")
		
			# Den ersten Waypoint mit allen Infos rauslesen
			gpxwpts=tagContentsOfXMLString(filecontentstring, '<wpt ', '</wpt>') # Das zieht die gesamte Beschreibung (inkl. Sub-Tags) aus dem GPX
			Say("Found %i Waypoints in GPX!", gpxwpts.length)

			if (gpxwpts.length) then
				# Den ersten (und wahrscheinlich einzigen) Waypoint auslesen
				gpxwpt=gpxwpts[0]
				Say("gpxwpt: %s", gpxwpt)
				# Diesen Waypoint weiter analysieren, zB Cache-ID (name) auslesen:
				cache_id=oneTagContentOfXMLString(gpxwpt, '<name>', '</name>')
				Say("cache_id: %s", cache_id)
				Say("cache_id (p): %p", cache_id) # Obwohl hier uU 0 steht, gehene Prüfungen gegen 0 und nil nicht!!
				cache_id=sprintf("%s", cache_id) # Mieser workaround!	
				#geht nid, is nil: Say("cache_id.length: %d", cache_id.length)
				if (cache_id.length<=0) then # Mieser workaround!
					cache_id="n/a"
				end if
#does not work	if (cache_id==nil) then
#does not work		Say("cache_id is nil")
#does not work	end if
#does not work	if (cache_id==null) then
#does not work		Say("cache_id is null")
#does not work	end if
#does not work	if (cache_id==0x0) then
#does not work		Say("cache_id is 0x0")
#does not work	end if
#does not work	if (!cache_id) then
#does not work		Say("cache_id is !")
#does not work	end if
#does not work	if (!cache_id or cache_id.length<=0) then
#does not work		Say("Setting cache_id with dummy value")
#does not work		cache_id="n/a"
#does not work	end if
				@dict.setObject_forKey(cache_id, "Cache-ID")

				# Diesen Waypoint weiter analysieren, zB Namen (desc) auslesen:
				name=oneTagContentOfXMLString(gpxwpt, '<desc>', '</desc>')
				if (!name) then
					name = gpxfilename
				end # if !name
				Say("name: %s", name)
				@dict.setObject_forKey(name, "Name")
				
				# Diesen Waypoint weiter analysieren, zB Type (type) auslesen:
				type=oneTagContentOfXMLString(gpxwpt, '<type>Geocache|', '</type>')
				Say("type: %s", type)
				if (!type) then
					type = "GPX"
				end # if !type
				@dict.setObject_forKey(type, "Type")

				# Diesen Waypoint weiter analysieren, zB Difficulty (groundspeak:difficulty) auslesen:
				difficulty=oneTagContentOfXMLString(gpxwpt, '<groundspeak:difficulty>', '</groundspeak:difficulty>')
				if (!difficulty) then
					difficulty = "n/a"
				end # if !difficulty
				Say("difficulty: %s", difficulty)
				@dict.setObject_forKey(difficulty, "Difficulty")

				# Diesen Waypoint weiter analysieren, zB Terrain (<groundspeak:terrain>) auslesen:
				terrain=oneTagContentOfXMLString(gpxwpt, '<groundspeak:terrain>', '</groundspeak:terrain>')
				if (!terrain) then
					terrain = "n/a"
				end # if !terrain
				Say("terrain: %s", terrain)
				@dict.setObject_forKey(terrain, "Terrain")

				# Diesen Waypoint weiter analysieren, zB URL ( <url>http://www.geocaching.com/seek/cache_details.aspx?guid=5b366bdd-629e-4086-a7eb-efebe27fc399</url>) auslesen:
				url=oneTagContentOfXMLString(gpxwpt, '<url>', '</url>')
				if (!url) then
					url = "n/a"
				end # if !url
				Say("url: %s", url)
				@dict.setObject_forKey(url, "URL")

				# Diesen Waypoint weiter analysieren, zB Size ( <groundspeak:container>Regular</groundspeak:container>) auslesen:
				size=oneTagContentOfXMLString(gpxwpt, '<groundspeak:container>', '</groundspeak:container>')
				if (!size) then
					size = "n/a"
				end # if !size
				Say("size: %s", size)
				@dict.setObject_forKey(size, "Size")
				
				# Country (<groundspeak:country>Germany</groundspeak:country>)
				country=oneTagContentOfXMLString(gpxwpt, '<groundspeak:country>', '</groundspeak:country>')
				if (!country) then
					country = "n/a"
				end # if !country
				Say("country: %s", country)
				@dict.setObject_forKey(country, "Country")

				# State (<groundspeak:state>Baden-Württemberg</groundspeak:state>)
				state=oneTagContentOfXMLString(gpxwpt, '<groundspeak:state>', '</groundspeak:state>')
				if (!state) then
					state = "n/a"
				end # if !state
				Say("state: %s", state)
				@dict.setObject_forKey(state, "State")

				# Diesen Waypoint weiter analysieren, zB Owner-Tag (<groundspeak:owner id="205158">benadar</groundspeak:owner>) auslesen:
				owner_tag=oneTagContentOfXMLString(gpxwpt, '<groundspeak:owner id=', '</groundspeak:owner>')
				
					# Diesen Waypoint weiter analysieren, zB Owner (<groundspeak:owner id="205158">benadar</groundspeak:owner>) auslesen:
					owner=oneTagContentOfXMLString(owner_tag, '>', '<')
					if (!owner) then
						owner = "n/a"
					end # if !owner
					Say("owner: %s", owner)
					@dict.setObject_forKey(owner, "Owner")

				# Diesen Waypoint weiter analysieren, zB Cache-Status-Details (<groundspeak:cache id="415023" available="True" archived="False" xmlns:groundspeak="http://www.groundspeak.com/cache/1/0">) auslesen:
				cache_status_tag=oneTagContentOfXMLString(gpxwpt, '<groundspeak:cache id=', '>')

					# Diesen Waypoint weiter analysieren, zB GroundSpeakCacheID (<groundspeak:cache id="415023" available="True" archived="False" xmlns:groundspeak="http://www.groundspeak.com/cache/1/0">) auslesen:
					groundspeakinternalcache_id=oneTagContentOfXMLString(cache_status_tag, '"', '"')
					if (!groundspeakinternalcache_id) then
						groundspeakinternalcache_id = "n/a"
					end # if !groundspeakinternalcache_id
					Say("groundspeakinternalcache_id: %s", groundspeakinternalcache_id)
					@dict.setObject_forKey(groundspeakinternalcache_id, "GroundSpeakInternalCache-ID")

					# Diesen Waypoint weiter analysieren, zB Available-Flag (<groundspeak:cache id="415023" available="True" archived="False" xmlns:groundspeak="http://www.groundspeak.com/cache/1/0">) auslesen:
					cache_status=oneTagContentOfXMLString(cache_status_tag, 'available="', '"')
					if (!cache_status) then
						cache_status = "n/a"
					end # if !cache_status
					Say("Status (Available): %s", cache_status)
					if (cache_status=="True") then
						@dict.setObject_forKey("Available", "Status")
					else
						@dict.setObject_forKey("Unavailable", "Status")
					end # if Active = True

					# Diesen Waypoint weiter analysieren, zB Archived-Flag (<groundspeak:cache id="415023" available="True" archived="False" xmlns:groundspeak="http://www.groundspeak.com/cache/1/0">) auslesen:
					cache_status=oneTagContentOfXMLString(cache_status_tag, 'archived="', '"')
					if (!cache_status) then
						cache_status = "n/a"
					end # if !cache_status
					Say("Status (Archived): %s", cache_status)
					if (cache_status=="True") then
						@dict.setObject_forKey("Archived", "Status")
					else
						# ignore, dann isser ja aktiv!! @dict.setObject_forKey("Unavailable", "Status")
					end # if Active = True

				# Diesen Waypoint weiter analysieren, zB Description (<groundspeak:long_description html="False">) auslesen:
				description_non_html=oneTagContentOfXMLString(gpxwpt, '<groundspeak:long_description html="False">', '</groundspeak:long_description>')
				if (description_non_html) then
					Say("description_non_html: %s", description_non_html)
					@dict.setObject_forKey(description_non_html, "Description")
				else
					Say("Unfortunately the GPX does not contain a non-HTML description")
					@dict.setObject_forKey("Undisplayable HTML stuff :-( ", "Description")
				end # if html desc
				
			else
				Say("No GPX-Waypoints found in GPX file")
			end # if keine Waypoints
		end # if kein Pocket Query

	return(@dict)
end # fillDict

# Liefert ein Array der Contents der Tags zurück
def tagContentsOfXMLString(xmlstring, beginTag, endTag)
	if (xmlstring) then
		# Split nach BeginTag
		contentsarray=xmlstring.split(beginTag)
		# Erstes Element ist ja VOR dem Tag (Es sei denn, das XML-File begann mit diesem Tag) -> Droppen!
		contentsarray.delete_at(0)
#		Say("mFound tag %i times.", contentsarray.length)

		# Isolieren des Tag-Contents (Abschneiden beim endTag)
		if (contentsarray.length) then
#			Say("mArray length is ok.", contentsarray.length)
			for i in (0...contentsarray.length)
#				Say("m i: %i", i)
				a=contentsarray[i].split(endTag)
#				Say("mFound endTag %i times.", a.length)
				contentsarray[i]=a[0]
			end # for index in array
		else
#			Say("mError: No tagBegin found!")
		end # if array.length != 0
	else
		contentsarray = Array.new
	end # if xmlstring

	return(contentsarray)
end # tagContentsOfXMLString

# Liefert nur das erste Vorkommen des Tags zurück!
def oneTagContentOfXMLString(xmlstring, beginTag, endTag)
	Say("Searching for \"" + beginTag + "\" to \"" + endTag + "\"");

	# Tags parsen
	array=tagContentsOfXMLString(xmlstring, beginTag, endTag)
	Say("nFound tag %i times.", array.length);
	
	# Nur die erste Occurency of Tag zurückliefern (alle anderen löschen)
	if (array.length > 2) then
		##falsch array.delete_at(0) # Erste Occurency ist eh nur das Gelumpe VOR dem Tag (es sei denn, das Tag ist das allererste im String, dann habe ich verschissen!)
		for i in (1...array.length) # ArrayEinträge NACH dem ersten einfach löschen
			array.delete_at(i)
		end # for
	end # if
	
	# Zurückliefern
#	Say("n-----------------------")
#	Say("nReturning THIS ONE TAG:")
#	Say(array[0])
#	Say("n-----------------------")
	return(array[0])
end # oneTagContentOfXMLString

def reload
	Say("CachesDataSource: reload!")
	self.removeAllObjects()	
	self.fillArray()	
	if (@_tableView) then
		@_tableView.reloadData
	else
		Say("@_tableView is nil! Can't reload data.")
	end
end # reload

# Initialization
def awakeFromNib
	# Init des Arrays und so
	Say("CachesDataSource: AwakeFromNib")
	@_array	= OSX::NSMutableArray.array.retain	# Diese Variablen nur via Accessor Methods lesen/schreiben. Danke :)
	@_lock	= OSX::NSLock.alloc.init.retain		# Diese Variablen nur via Accessor Methods lesen/schreiben. Danke :)
	#@_tableView = nil							# Diese Variablen nur via Accessor Methods lesen/schreiben. Danke :)
	#_debug		= YES

	# Datasource fuellen und TV reloaden
	self.reload()
	
end # def awakeFromNib

# Kleines Helferlein, damit ich nicht auf das Array direkt zugreifen muss.
def count()
#	return (@_array.count)
	return (self.array.count)
end # def count

def array()
	return @_array
end # def array


# lock und unlock werden immer von dem DataSource-Benutzenden (Thread/Classe) aufgerufen! 
#  Es gibt nur eine Ausnahme: tableView:objectFor...
def lock()
	Say("lock! (stock and two smoking barrels)")
	if (@_lock) then
		@_lock.lock
	else
		Say("lock was not initialized. Initializing.")
		@_lock	= OSX::NSLock.alloc.init.retain
		@_lock.lock
	end
	Say("lock done!")
end # def lock

# lock und unlock werden immer von dem DataSource-Benutzenden (Thread/Classe) aufgerufen! 
# Es gibt nur eine Ausnahme: tableView:objectFor... */
def unlock()
	Say("unlock!")
	@_lock.unlock
	Say("unlock done!")
end # def lock

# Das wird von tableView_objectValueForTableColumn_row und von evtl. externen Callern benötigt
def objectValueForTableColumnIdentifier_row(tableColumnIdentifier, tableRow)
	@result="what?"

	# RowDict rausholen, ggf. raus hier!
	@rowDict		=	self.objectAtIndex(tableRow)
	if !@rowDict then
		Say("RowDict is nil for row %i.", tableRow)
		return ("no object")
	end # if rowDict

	# Daumen draufhalten
	@rowDict.retain()

	# Den Cell-Inhalt regeln
	Say("tableColumnIdentifier: %s", tableColumnIdentifier.to_s)
	#Say("@tableColumnIdentifier: %s", @tableColumnIdentifier.to_s)
	@result = @rowDict.objectForKey(tableColumnIdentifier)
	Say("@result: %s", @result)
	
	# Den Daumen wieder runternehmen
	@rowDict.release()

	return (@result)
end # def objectValueForTableColumnIdentifier_row


# Gehört zum TableView-Datasource: Holt das Object für den Service aus dem Array. */
#def tableView:(NSTableView *)theTableView objectValueForTableColumn:(NSTableColumn *)theColumn row:(int)row 
def tableView_objectValueForTableColumn_row(tableView, tableColumn, tableRow) # KEIN trailing "_"!!
	@result=nil
	Say("tableView_objectValueForTableColumn_row: %i and column %p.", tableRow, tableColumn)

	# TableColumn checken, ggf. raus hier!
	if !tableColumn then
		Say("tableColumn is nil for row %i.", tableRow)
		return (nil)
	end # if rowDict

	# locken!
	self.lock()

		# Value holen
		@tableColumnIdentifier = tableColumn.identifier()
		result=objectValueForTableColumnIdentifier_row_(@tableColumnIdentifier, tableRow)

 	# unlocken! 
	self.unlock()

	# Zurück zu Lück!
	Say("result: %s", @result.to_s)
	return (@result)
	#return(OSX::NSString.stringWithString_("Foo"))
end # def tableView_objectValueForTableColumn_row_

def numberOfRowsInTableView(theTableView)
#	Say("numberOfRowsInTableView")
	if (self.array) then
		result = self.array.count
	else
		result = 0
	end
	Say("numberOfRowsInTableView: %i", result)
	return (result)
end # def numberOfRowsInTableView


def removeAllObjects
	Say("removeAllObjects")
#DAS locken MACHT AUFRUFER	[self lock];
	if (self.array) then
		self.array.removeAllObjects()
	end
#könnte zu Deadlock führen!! 2006-06-01: AUS!	[receivedFilesTableView reloadData];
	# locken! #DAS MACHT AUFRUFER	[self unlock];
end # def removeAllObjects

#    tableView_setObjectValue_forTableColumn_row_
def tableView_setObjectValue_forTableColumn_row(tableView, objectValue, tableColum, tableRow)
	Say("setObjectValue is not used and therefore not implemented")
	return (nil)
end # def setObjectValue

def objectAtIndex(index)
	Say("objectAtIndex %i", index)
	
	# Sichergehen, dass hier immer nur ein Task sich ein Dict holt! und damit fuhrwerkt! 
#	[lock lock];
	
	# Der eigentliche job
	result=self.array.objectAtIndex(index)

#	[lock unlock];
	
	return(result)
end # def objectAtIndex


def objectEnumerator
	Say("objectEnumerator"); 
	return(self.array.objectEnumerator)
end # def objectEnumerator

def tableView_sortDescriptorsDidChange_(tableView, oldDescriptors)
	Say("tableView_sortDescriptorsDidChange")
end # def tableView_sortDescriptorsDidChange

def tableView_writeRows_toPasteboard_(tableView, rows, pboard)
	Say("Huhu") 
	return (NO); # Man kann nicht aus der TV rausdraggen! - Vorerst.
end # def tableView_writeRows_toPasteboard_

def addObject(anObject)
	Say("addObject start") 
	self.array.addObject(anObject)
# NEIN -> Das führt zu Locking problem, weil die "ObjectValueFor..." auch locked!!!!	
#	[receivedFilesTableView reloadData];
#
	Say("addObject end")
end # def addObject

def setTableView(tableView)
	Say("setTableView %p", tableView) 
	@_tableView=tableView
end # def setTableView

def tableView
	Say("tableView %p", @_tableView) 
	return(@_tableView)
end # def tableView



#def Say(*args)
#    if (DEBUGCACHESDATASOURCE==NO) then return # Debugging ist aus -> raus hier
#	else
#		OSX::NSLog(*args)
#	end # else !debug
#end # Say


end # class CachesDataSource
