#!/usr/bin/perl
#
# File     : gpxtracktoexcel.pl
# Author   : Heiko Kretschmer
# Purpose  : This script reads GPX format tracklogs and converts them in csv (comma separated viles :) ). 
# History & ToDos & Bugs: see History.h
#
# Created by Kretschmer, Heiko on 10.08.08.
# Copyright 2008 Heiko Kretschmer. All rights reserved.
#

###########################################
# Modules                                 #
###########################################
use strict;

###########################################
# Globals, später mal Params              #
###########################################
#my $maximumRegularIssueIndex="10000";

###########################################
# Constants                               #
###########################################
#use constant LONGREADLEN => 2000000; # 2.000.000 Bytes == 2 MB


###########################################
# Main                                    #
###########################################

# Übergabeparameter lesen
my $inputFilePath=shift;  # Infile
my $outputFilePath=shift; # Outfile
my $dividerDecimal=shift;  # Trenner, im deutschen also ein "Komma" wie in "23,5", in english wäre das "23.5".

# Parametercheck
if ($inputFilePath eq "" || $outputFilePath eq "" || $dividerDecimal eq "") { die ("Faehlaeaeaeaeaerrrr!"); }
print ("Processing '$inputFilePath' -> '$outputFilePath' with '$dividerDecimal'.\n");

# File öffnen
if (open(fi, "<$inputFilePath"))
{
	# Alle Zeilen eine Variable schreiben, das könnte zwar RAM fressen ist aber super-easy zu parsen
	my	$gpxContent="";
	my	$lineCount=0;
	while (<fi>)
	{
		# Noch ne Zeile
		$lineCount++;
		#print("$_\nfoo\n");
		$gpxContent=$gpxContent . $_;
	}
	close(fi);
	print ("Read $lineCount lines.\n");

	# Output-File-Pfad, Tilde expanden
# perl haengt hier:
#    $outputFilePath =~ s{
#      ^~             # find a leading tilde
#      (               # save this in $1
#          [^/]        # a non-slash character
#          *           # repeated 0 or more times (0 means me)
#      )
#      }{
#          $1
#          ? (getpwnam($1))[7]
#          : ( $ENV{HOME} || $ENV{LOGDIR} )
#    }ex;
#die ("WEG!");

	# Nun das gelesene parsen und gleich in das Outfile schreiben!
	if (open(fo, "> $outputFilePath"))
	{
		print ("Reading trackpoints...\n");
		# Nun den Inhalt des GPX parsen...
		my @trackPoints=getXMLTagContentFromString($gpxContent, "<trkpt ", "</trkpt>");
		print ("Read $#trackPoints trackpoints.\n");
		# Nun die Trackpoints zerlegen
		my	$trackPointString;
		print (fo "\"" . "Date" . "\"\t\"" . "Time" . "\"\t\"" . "Lat" . "\"\t\"" . "Lon" . "\"\t\"" . "Ele" . "\"\t\"" . "Atemp" . "\"\n");
		#		print (fo "\"" . "Time" . "\",\t\"" . "Lat" . "\",\t\"" . "Lon" . "\",\t\"" . "Ele" . "\",\t\"" . "Atemp" . " \"\n");
		foreach $trackPointString (@trackPoints)
		{
			# Longitude rausholen
			my @lat	=getXMLTagContentFromString($trackPointString, "lat=\"", "\" ");
			my @lon	=getXMLTagContentFromString($trackPointString, " lon=\"", "\">");
			my @ele	=getXMLTagContentFromString($trackPointString, "<ele>", "</ele>");
			my @datetime=getXMLTagContentFromString($trackPointString, "<time>", "</time>");
			my @datetimecomp = split /T/, @datetime[0]; # Das Datum wird durch ein T von der Zeit getrennt. Nach der Time kommt die Timezone (Z)
			my $date=@datetimecomp[0];
			my @timecomp=split /Z/, @datetimecomp[1]; # Nach der Time kommt die Timezone (Z), Timezone abschneiden
			my $time = @timecomp[0];
			my @atemp=getXMLTagContentFromString($trackPointString, "<gpxtpx:atemp>", "</gpxtpx:atemp>");
			# Wemma ein Deutsches Excel haben, muessen wir die verfickten Kommas benutzen...
			#	my	($etime, $elat, $elon, $eele, $eatemp);
			if (1 == 1)
			{
				#@time[0]	=	convertNumberFormat(@time[0]);
				@lat[0]		=	convertNumberFormat(@lat[0]);
				@lon[0]		=	convertNumberFormat(@lon[0]);
				@ele[0]		=	convertNumberFormat(@ele[0]);
				@atemp[0]	=	convertNumberFormat(@atemp[0]);
				#				$etime	=	convertNumberFormat(@time[0]);
				#				$elat	=	convertNumberFormat(@lat[0]);
				#				$elon	=	convertNumberFormat(@lon[0]);
				#				$eele	=	convertNumberFormat(@ele[0]);
				#				$eatemp	=	convertNumberFormat(@atemp[0]);
			}
			# Das ganze nun ins csv schreiben
			#"	my $outputLine="\"" . @time[0] . "\",\t\"" . @lat[0] . "\",\t\"" . @lon[0] . "\",\t\"" . @ele[0] . "\",\t\"+" . @atemp[0] . " \"\n";
			my $outputLine=$date . "\t" . $time . "\t" . @lat[0] . "\t" . @lon[0] . "\t" . @ele[0] . "\t" . @atemp[0] . "\n";
			#			my $outputLine="\"" . $etime . "\"\t" . $elat . "\t" . $elon . "\t" . $eele . "\t" . $eatemp . "\n";
			print (fo $outputLine);
		}
	}
	else
	{
		print("Output file '$outputFilePath' could not be opened :-(\n");
	}
	close(fo);
	print("Conversion done.\n");
}
else
{
	die ("Could not open file '$inputFilePath' for reading.\n");
}


# File öffnen, lese in ein Array
#if (open(f, "<$inputfFilePath"))
#{
#	my @lines = readline(f);
#	#deb
#	@lines[1] = "NOCHEINELEMENT";
#	print ("read $#lines lines\n");
#	print ("lines: @lines[0]\n");
#	if ($#lines > 0)
#	{ # Wenn mehr als eine Zeile gelesen wurde, alle Elemente aneinanderbeppen
#	}
#	close(f);
#}
#else
#{
#	die ("Could not open file '$inputfFilePath'.\n");
#}

#die ("ferdig");

#my @pstys=getXMLTagContentFromString($content, "<psty ", "/>");
#			my @pstynams;
#			foreach $psty (@pstys)
#			{
#				my $pstynam=getXMLTagComponentFromXMLTagContent($psty, "pnam=\"c_", "\"");
#				push (@pstynams, $pstynam); # Das Paragraph-Style-Name-Array fuellen
#			}


###########################################
# Functions                               #
###########################################

# "nn.mm" zu "nn,mm" konvertieren
sub convertNumberFormat()
{
	my $in = shift;
	my $out="";
	
	# Konvertieren von en zu dt, das dt. Excel will "," statt "." als Dezimaltrenner
	$_ = $in;
	s/\./$dividerDecimal/;
	#s/\./,/;
	$out = $_;
	
	return($out);
}

# Diese Methode filtert mir den Inhalt eines XML-Tags raus.
sub getXMLTagContentFromString()
{
	my	$inputString=shift;
	my	$tagBegin=shift;
	my	$tagEnd=shift;
	my	@result;
	my	$tagString;
	my	$t;

	#	print ("tagBegin: '$tagBegin'\n");
	#	print ("tagEnd  : '$tagEnd'\n");

	# Zuerstmal anhand der gefundenen Begin-Tags zerlegen
	my	@tagStrings = split /$tagBegin/, $inputString;
	
	# Der erste SubString ist ggf. auch das Gelumpe VOR dem ersten Tagvorkommen. Recht wahrscheinlich sogar. Also gleich mal blind wegwerfen.
	shift (@tagStrings); # removes first element of array (see also pop, push, unshift @ http://www.tizag.com/perlT/perlarrays.php)
	#	print ("$#tagStrings+1 tagStrings found.\n");
	
	# Die gefundenen SubStrings durchkaspern
	my $i=0;
	foreach $tagString (@tagStrings)
	{
		# Checken, ob in diesem String ein tagEnd ist. Wenn ja, abschneiden...
		my @tagStringComp = split /$tagEnd/, $tagString;
		my $tagStringComp = @tagStringComp[0];
		push(@result,  $tagStringComp);
		#		print ("comp $i: '$tagStringComp'\n");
		$i++;
	}

	# Deb
	#die
	
	# Zurück zu Luc!
	return(@result);
}

